@@ -5,7 +5,7 @@
  * 2.0 (the "License"); you may not use this file except in compliance with the
  * License. You may obtain a copy of the License at:
  *
- * http://www.apache.org/licenses/LICENSE-2.0
+ * https://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
@@ -34,15 +34,14 @@ import io.netty.channel.local.LocalServerChannel;
 import io.netty.handler.ssl.util.InsecureTrustManagerFactory;
 import io.netty.handler.ssl.util.SelfSignedCertificate;
 import io.netty.util.ReferenceCountUtil;
+import io.netty.util.concurrent.Future;
+import io.netty.util.concurrent.ImmediateEventExecutor;
 import io.netty.util.concurrent.Promise;
-import org.hamcrest.Matchers;
-import org.junit.AfterClass;
-import org.junit.Assume;
-import org.junit.BeforeClass;
-import org.junit.Test;
-import org.junit.runner.RunWith;
-import org.junit.runners.Parameterized;
-import org.junit.runners.Parameterized.Parameters;
+import io.netty.util.internal.ThreadLocalRandom;
+import org.junit.jupiter.api.AfterAll;
+import org.junit.jupiter.api.BeforeAll;
+import org.junit.jupiter.params.ParameterizedTest;
+import org.junit.jupiter.params.provider.MethodSource;
 
 import javax.net.ssl.KeyManagerFactory;
 import javax.net.ssl.SSLContext;
@@ -65,30 +64,36 @@ import java.util.concurrent.ThreadFactory;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicBoolean;
 
-import static org.junit.Assert.assertEquals;
-import static org.junit.Assert.assertNotEquals;
-import static org.junit.Assert.assertNotNull;
-import static org.junit.Assert.assertThat;
-import static org.junit.Assert.assertTrue;
+import static io.netty.handler.ssl.OpenSslTestUtils.checkShouldUseKeyManagerFactory;
+import static org.junit.jupiter.api.Assertions.assertEquals;
+import static org.junit.jupiter.api.Assertions.assertInstanceOf;
+import static org.junit.jupiter.api.Assertions.assertNotNull;
+import static org.junit.jupiter.api.Assertions.assertTrue;
+import static org.junit.jupiter.api.Assumptions.assumeTrue;
 
-@RunWith(Parameterized.class)
 public class OpenSslPrivateKeyMethodTest {
     private static final String RFC_CIPHER_NAME = "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256";
     private static EventLoopGroup GROUP;
     private static SelfSignedCertificate CERT;
-    private static ExecutorService EXECUTOR;
+    private static DelayingExecutor EXECUTOR;
 
-    @Parameters(name = "{index}: delegate = {0}")
-    public static Collection<Object[]> parameters() {
+    static Collection<Object[]> parameters() {
         List<Object[]> dst = new ArrayList<Object[]>();
-        dst.add(new Object[] { true });
-        dst.add(new Object[] { false });
+        for (int a = 0; a < 2; a++) {
+            for (int b = 0; b < 2; b++) {
+                for (int c = 0; c < 2; c++) {
+                    dst.add(new Object[] { a == 0, b == 0, c == 0 });
+                }
+            }
+        }
         return dst;
     }
 
-    @BeforeClass
+    @BeforeAll
     public static void init() throws Exception {
-        Assume.assumeTrue(OpenSsl.isBoringSSL());
+        checkShouldUseKeyManagerFactory();
+
+        assumeTrue(OpenSsl.isBoringSSL());
         // Check if the cipher is supported at all which may not be the case for various JDK versions and OpenSSL API
         // implementations.
         assumeCipherAvailable(SslProvider.OPENSSL);
@@ -96,7 +101,7 @@ public class OpenSslPrivateKeyMethodTest {
 
         GROUP = new DefaultEventLoopGroup();
         CERT = new SelfSignedCertificate();
-        EXECUTOR = Executors.newCachedThreadPool(new ThreadFactory() {
+        EXECUTOR = new DelayingExecutor(new ThreadFactory() {
             @Override
             public Thread newThread(Runnable r) {
                 return new DelegateThread(r);
@@ -104,7 +109,7 @@ public class OpenSslPrivateKeyMethodTest {
         });
     }
 
-    @AfterClass
+    @AfterAll
     public static void destroy() {
         if (OpenSsl.isBoringSSL()) {
             GROUP.shutdownGracefully();
@@ -113,26 +118,20 @@ public class OpenSslPrivateKeyMethodTest {
         }
     }
 
-    private final boolean delegate;
-
-    public OpenSslPrivateKeyMethodTest(boolean delegate) {
-        this.delegate = delegate;
-    }
-
     private static void assumeCipherAvailable(SslProvider provider) throws NoSuchAlgorithmException {
         boolean cipherSupported = false;
         if (provider == SslProvider.JDK) {
             SSLEngine engine = SSLContext.getDefault().createSSLEngine();
             for (String c: engine.getSupportedCipherSuites()) {
-               if (RFC_CIPHER_NAME.equals(c)) {
-                   cipherSupported = true;
-                   break;
-               }
+                if (RFC_CIPHER_NAME.equals(c)) {
+                    cipherSupported = true;
+                    break;
+                }
             }
         } else {
             cipherSupported = OpenSsl.isCipherSuiteAvailable(RFC_CIPHER_NAME);
         }
-        Assume.assumeTrue("Unsupported cipher: " + RFC_CIPHER_NAME, cipherSupported);
+        assumeTrue(cipherSupported, "Unsupported cipher: " + RFC_CIPHER_NAME);
     }
 
     private static SslHandler newSslHandler(SslContext sslCtx, ByteBufAllocator allocator, Executor executor) {
@@ -148,15 +147,13 @@ public class OpenSslPrivateKeyMethodTest {
 
         final KeyManagerFactory kmf = OpenSslX509KeyManagerFactory.newKeyless(CERT.cert());
 
-        final SslContext sslServerContext = SslContextBuilder.forServer(kmf)
+        return SslContextBuilder.forServer(kmf)
                 .sslProvider(SslProvider.OPENSSL)
                 .ciphers(ciphers)
                 // As this is not a TLSv1.3 cipher we should ensure we talk something else.
-                .protocols(SslUtils.PROTOCOL_TLS_V1_2)
+                .protocols(SslProtocols.TLS_v1_2)
+                .option(OpenSslContextOption.PRIVATE_KEY_METHOD, method)
                 .build();
-
-        ((OpenSslContext) sslServerContext).setPrivateKeyMethod(method);
-        return sslServerContext;
     }
 
     private SslContext buildClientContext()  throws Exception {
@@ -164,31 +161,43 @@ public class OpenSslPrivateKeyMethodTest {
                 .sslProvider(SslProvider.JDK)
                 .ciphers(Collections.singletonList(RFC_CIPHER_NAME))
                 // As this is not a TLSv1.3 cipher we should ensure we talk something else.
-                .protocols(SslUtils.PROTOCOL_TLS_V1_2)
+                .protocols(SslProtocols.TLS_v1_2)
                 .trustManager(InsecureTrustManagerFactory.INSTANCE)
                 .build();
     }
 
-    private Executor delegateExecutor() {
-       return delegate ? EXECUTOR : null;
+    private static Executor delegateExecutor(boolean delegate) {
+        return delegate ? EXECUTOR : null;
+    }
+    private SslContext buildServerContext(OpenSslAsyncPrivateKeyMethod method) throws Exception {
+        List<String> ciphers = Collections.singletonList(RFC_CIPHER_NAME);
+
+        final KeyManagerFactory kmf = OpenSslX509KeyManagerFactory.newKeyless(CERT.cert());
+
+        return SslContextBuilder.forServer(kmf)
+                .sslProvider(SslProvider.OPENSSL)
+                .ciphers(ciphers)
+                // As this is not a TLSv1.3 cipher we should ensure we talk something else.
+                .protocols(SslProtocols.TLS_v1_2)
+                .option(OpenSslContextOption.ASYNC_PRIVATE_KEY_METHOD, method)
+                .build();
     }
 
-    private void assertThread() {
+    private static void assertThread(boolean delegate) {
         if (delegate && OpenSslContext.USE_TASKS) {
             assertEquals(DelegateThread.class, Thread.currentThread().getClass());
-        } else {
-            assertNotEquals(DelegateThread.class, Thread.currentThread().getClass());
         }
     }
 
-    @Test
-    public void testPrivateKeyMethod() throws Exception {
+    @ParameterizedTest(name = "{index}: delegate = {0}, async = {1}, newThread={2}")
+    @MethodSource("parameters")
+    public void testPrivateKeyMethod(final boolean delegate, boolean async, boolean newThread) throws Exception {
         final AtomicBoolean signCalled = new AtomicBoolean();
-        final SslContext sslServerContext = buildServerContext(new OpenSslPrivateKeyMethod() {
+        OpenSslPrivateKeyMethod keyMethod = new OpenSslPrivateKeyMethod() {
             @Override
             public byte[] sign(SSLEngine engine, int signatureAlgorithm, byte[] input) throws Exception {
                 signCalled.set(true);
-                assertThread();
+                assertThread(delegate);
 
                 assertEquals(CERT.cert().getPublicKey(),
                         engine.getSession().getLocalCertificates()[0].getPublicKey());
@@ -214,7 +223,10 @@ public class OpenSslPrivateKeyMethodTest {
             public byte[] decrypt(SSLEngine engine, byte[] input) {
                 throw new UnsupportedOperationException();
             }
-        });
+        };
+
+        final SslContext sslServerContext = async ? buildServerContext(
+                new OpenSslPrivateKeyMethodAdapter(keyMethod, newThread)) : buildServerContext(keyMethod);
 
         final SslContext sslClientContext = buildClientContext();
         try {
@@ -226,7 +238,7 @@ public class OpenSslPrivateKeyMethodTest {
                     @Override
                     protected void initChannel(Channel ch) {
                         ChannelPipeline pipeline = ch.pipeline();
-                        pipeline.addLast(newSslHandler(sslServerContext, ch.alloc(), delegateExecutor()));
+                        pipeline.addLast(newSslHandler(sslServerContext, ch.alloc(), delegateExecutor(delegate)));
 
                         pipeline.addLast(new SimpleChannelInboundHandler<Object>() {
                             @Override
@@ -262,7 +274,7 @@ public class OpenSslPrivateKeyMethodTest {
                         @Override
                         protected void initChannel(Channel ch) {
                             ChannelPipeline pipeline = ch.pipeline();
-                            pipeline.addLast(newSslHandler(sslClientContext, ch.alloc(), delegateExecutor()));
+                            pipeline.addLast(newSslHandler(sslClientContext, ch.alloc(), delegateExecutor(delegate)));
 
                             pipeline.addLast(new SimpleChannelInboundHandler<Object>() {
                                 @Override
@@ -290,10 +302,10 @@ public class OpenSslPrivateKeyMethodTest {
                     Channel client = client(server, clientHandler);
                     try {
                         client.writeAndFlush(Unpooled.wrappedBuffer(new byte[] {'P', 'I', 'N', 'G'}))
-                              .syncUninterruptibly();
+                                .syncUninterruptibly();
 
-                        assertTrue("client timeout", clientPromise.await(5L, TimeUnit.SECONDS));
-                        assertTrue("server timeout", serverPromise.await(5L, TimeUnit.SECONDS));
+                        assertTrue(clientPromise.await(5L, TimeUnit.SECONDS), "client timeout");
+                        assertTrue(serverPromise.await(5L, TimeUnit.SECONDS), "server timeout");
 
                         clientPromise.sync();
                         serverPromise.sync();
@@ -312,12 +324,26 @@ public class OpenSslPrivateKeyMethodTest {
         }
     }
 
-    @Test
-    public void testPrivateKeyMethodFails() throws Exception {
+    @ParameterizedTest(name = "{index}: delegate = {0}")
+    @MethodSource("parameters")
+    public void testPrivateKeyMethodFailsBecauseOfException(final boolean delegate) throws Exception {
+        testPrivateKeyMethodFails(delegate, false);
+    }
+
+    @ParameterizedTest(name = "{index}: delegate = {0}")
+    @MethodSource("parameters")
+    public void testPrivateKeyMethodFailsBecauseOfNull(final boolean delegate) throws Exception {
+        testPrivateKeyMethodFails(delegate, true);
+    }
+
+    private void testPrivateKeyMethodFails(final boolean delegate, final boolean returnNull) throws Exception {
         final SslContext sslServerContext = buildServerContext(new OpenSslPrivateKeyMethod() {
             @Override
             public byte[] sign(SSLEngine engine, int signatureAlgorithm, byte[] input) throws Exception {
-                assertThread();
+                assertThread(delegate);
+                if (returnNull) {
+                    return null;
+                }
                 throw new SignatureException();
             }
 
@@ -329,9 +355,9 @@ public class OpenSslPrivateKeyMethodTest {
         final SslContext sslClientContext = buildClientContext();
 
         SslHandler serverSslHandler = newSslHandler(
-                sslServerContext, UnpooledByteBufAllocator.DEFAULT, delegateExecutor());
+                sslServerContext, UnpooledByteBufAllocator.DEFAULT, delegateExecutor(delegate));
         SslHandler clientSslHandler = newSslHandler(
-                sslClientContext, UnpooledByteBufAllocator.DEFAULT, delegateExecutor());
+                sslClientContext, UnpooledByteBufAllocator.DEFAULT, delegateExecutor(delegate));
 
         try {
             try {
@@ -345,7 +371,7 @@ public class OpenSslPrivateKeyMethodTest {
                         Throwable clientCause = clientSslHandler.handshakeFuture().await().cause();
                         Throwable serverCause = serverSslHandler.handshakeFuture().await().cause();
                         assertNotNull(clientCause);
-                        assertThat(serverCause, Matchers.instanceOf(SSLHandshakeException.class));
+                        assertInstanceOf(SSLHandshakeException.class, serverCause);
                     } finally {
                         client.close().sync();
                     }
@@ -385,4 +411,70 @@ public class OpenSslPrivateKeyMethodTest {
             super(target);
         }
     }
+
+    private static final class OpenSslPrivateKeyMethodAdapter implements OpenSslAsyncPrivateKeyMethod {
+        private final OpenSslPrivateKeyMethod keyMethod;
+        private final boolean newThread;
+
+        OpenSslPrivateKeyMethodAdapter(OpenSslPrivateKeyMethod keyMethod, boolean newThread) {
+            this.keyMethod = keyMethod;
+            this.newThread = newThread;
+        }
+
+        @Override
+        public Future<byte[]> sign(final SSLEngine engine, final int signatureAlgorithm, final byte[] input) {
+            final Promise<byte[]> promise = ImmediateEventExecutor.INSTANCE.newPromise();
+            try {
+                if (newThread) {
+                    // Let's run these in an extra thread to ensure that this would also work if the promise is
+                    // notified later.
+                    new DelegateThread(new Runnable() {
+                        @Override
+                        public void run() {
+                            try {
+                                // Let's sleep for some time to ensure we would notify in an async fashion
+                                Thread.sleep(ThreadLocalRandom.current().nextLong(100, 500));
+                                promise.setSuccess(keyMethod.sign(engine, signatureAlgorithm, input));
+                            } catch (Throwable cause) {
+                                promise.setFailure(cause);
+                            }
+                        }
+                    }).start();
+                } else {
+                    promise.setSuccess(keyMethod.sign(engine, signatureAlgorithm, input));
+                }
+            } catch (Throwable cause) {
+                promise.setFailure(cause);
+            }
+            return promise;
+        }
+
+        @Override
+        public Future<byte[]> decrypt(final SSLEngine engine, final byte[] input) {
+            final Promise<byte[]> promise = ImmediateEventExecutor.INSTANCE.newPromise();
+            try {
+                if (newThread) {
+                    // Let's run these in an extra thread to ensure that this would also work if the promise is
+                    // notified later.
+                    new DelegateThread(new Runnable() {
+                        @Override
+                        public void run() {
+                            try {
+                                // Let's sleep for some time to ensure we would notify in an async fashion
+                                Thread.sleep(ThreadLocalRandom.current().nextLong(100, 500));
+                                promise.setSuccess(keyMethod.decrypt(engine, input));
+                            } catch (Throwable cause) {
+                                promise.setFailure(cause);
+                            }
+                        }
+                    }).start();
+                } else {
+                    promise.setSuccess(keyMethod.decrypt(engine, input));
+                }
+            } catch (Throwable cause) {
+                promise.setFailure(cause);
+            }
+            return promise;
+        }
+    }
 }
